#include "Transformer.hpp"

using namespace Render;

void IWindowTransformer::preWindowRender(CSurfacePassElement::SRenderData* pRenderData) {
    ;
}

void IWindowTransformer::amendTransformedRenderData(const CBox& currentBox, SMotionBlurData* pMotionBlurData) {
    ;
}
