#pragma once

#include "includes.hpp"
#include "debug/log/Logger.hpp"
#include "helpers/Color.hpp"
#include "macros.hpp"
#include "desktop/DesktopTypes.hpp"

#if !defined(__GXX_RTTI)
#error "Hyprland requires C++ RTTI. Shit will hit the fan otherwise. Do not even try."
#endif
